Attribute VB_Name = "Module1"
'===========================================================================
' Subject: CALCULATE PI                      Date: 01-05-97 (07:59)
' Author:  Jason Stratos Papadopoulos        Code: QB, QBasic, PDS
' Origin:  comp.lang.basic.misc            Packet: ALGOR.ABC
'===========================================================================
'DECLARE SUB PrintOut (sum%(), words%)
'DECLARE SUB Multiply (term%(), words%, mult&, firstword%)
'DECLARE SUB Divide (term%(), words%, denom&, firstword%)
'DECLARE SUB Add (sum%(), term%(), words%, sign%, firstword%)
'DECLARE SUB FastDivide (term%(), words%, denom&)

'Program to calculate pi, version 2.0
'The algorithm used is Gregory's series with Euler acceleration.
'This program uses the optimal Euler 2/3 rule: rather than use Euler's
'series for all the terms, compute instead 1/3 of the terms using
'Gregory's series and the rest using Euler's. It can be shown that
'each term in this compound series cuts the error by a factor of 3,
'while using only Euler's series has each term cut the error by a
'factor of 2. This is a major timesaver: it reduces the number of terms
'to be added up by over 35%, and of the terms that remain 1/3 can
'be crunched out faster than normal! The code also includes some tricks
'to speed things up (like reducing the size of the arrays Euler's series
'works on).
'
'Converging faster also means more digits can be computed. Some tests
'show the program is capable of computing about 51,000 digits of pi,
'and is quite fast if compiled (5000 digits in about 90 seconds on
'a 486 66MHz computer). I'd be grateful if someone can help me code
'the Divide and FastDivide SUBs in assembly, which can probably make
'the program twice as fast. Comments or questions to jasonp@wam.umd.edu
'
'Converted to Visual Basic by Claus von der Burchard


Option Explicit
'-----------------------------------------------------------------------

Public Function MPi(Digits As Long) As String
    Dim Words As Long
    Dim terms As Long
    Dim denom As Long
    Dim sign As Long
    Dim x As Long
    Dim firstword As Long
    
    Dim sum() As Long
    Dim term() As Long
    
    Words = Digits& \ 4 + 4
    terms& = CLng(Digits& / 0.477) \ 3 + 1
    If terms& Mod 2 > 0 Then terms& = terms& + 1
    
    ReDim sum(Words), term(Words)
                                              'Gregory's Series-------
    sum(1) = 1: denom& = 3: sign = -1
    
    For x& = 1 To terms& - 1
      
       Call FastDivide(term(), Words, denom&)
       Call Add(sum(), term(), Words, sign, 2)
       denom& = denom& + 2: sign = -sign
    
    Next x&
                                              'Euler's Acceleration---
    firstword = 2: x& = 1
    Call FastDivide(term(), Words, 2 * denom&)
    
    Do Until firstword = Words
    
       denom& = denom& + 2
       Call Add(sum(), term(), Words, sign, firstword)
       Call Divide(term(), Words, denom&, firstword)
       Call Multiply(term(), Words, x&, firstword)
      
       If term(firstword) = 0 Then firstword = firstword + 1
       x& = x& + 1
    
    Loop
                                              'Finish up--------------
    Call Add(sum(), term(), Words, sign, firstword)
    Call Multiply(sum(), Words, 4, 1)
    MPi = PrintOut(sum(), Words)
End Function

'--------------------------------------------------------------------
Sub Add(sum() As Long, term() As Long, Words As Long, sign As Long, firstword As Long)
    Dim x As Long
    
    If sign = 1 Then
      
       'add it on
       For x = Words To firstword Step -1
          sum(x) = sum(x) + term(x)
          If sum(x) >= 10000 Then
             sum(x - 1) = sum(x - 1) + 1
             sum(x) = sum(x) - 10000
          End If
       Next x
    
    Else
    
        'subtract it off
        For x = Words To firstword Step -1
           sum(x) = sum(x) - term(x)
           If sum(x) < 0 Then
              sum(x - 1) = sum(x - 1) - 1
              sum(x) = sum(x) + 10000
           End If
        Next x
    
    End If
End Sub

'-------------------------------------------------------------------
Sub Divide(term() As Long, Words As Long, denom As Long, firstword As Long)
    Dim x As Long
    Dim dividend As Long
    Dim quotient As Long
    Dim remainder As Long
    
    For x = firstword To Words
       dividend& = remainder& * 10000 + term(x)
       quotient = dividend& \ denom&
       term(x) = quotient
       remainder& = dividend& - quotient * denom&
    Next x

End Sub

'------------------------------------------------------------------------
Sub FastDivide(term() As Long, Words As Long, denom As Long)
    'not really a fast divide, but there are fewer operations
    'since dividend& below doesn't have term(x) added on (always 0)
    Dim remainder As Long
    Dim dividend As Long
    Dim quotient As Long
    Dim x As Long
    
    remainder& = 1
    For x = 2 To Words
       dividend& = remainder& * 10000
       quotient = dividend& \ denom&
       term(x) = quotient
       remainder& = dividend& - quotient * denom&
    Next x
    
End Sub

'---------------------------------------------------------------------
Sub Multiply(term() As Long, Words As Long, mult As Long, firstword As Long)
    Dim x As Long
    Dim product As Long
    Dim carry As Long
    
    For x = Words To firstword Step -1
       product& = mult& * term(x) + carry&
       term(x) = product& Mod 10000
       carry& = (product& - term(x)) \ 10000
    Next x

End Sub

'------------------------------------------------------------------
Private Function PrintOut(sum() As Long, Words As Long, Optional Delimiter As String = "") As String
    Dim I As Long, j As Long
    PrintOut = "3,"
    I = 2
    Do Until I = Words - 1
       j = sum(I)
       If j > 999 Then
           PrintOut = PrintOut & Delimiter & Right$(Str$(j), 4)
       ElseIf j > 99 Then
           PrintOut = PrintOut & Delimiter & Right$(Str$(j), 3)
       ElseIf j > 9 Then
           PrintOut = PrintOut & Delimiter & Right$(Str$(j), 2)
       Else
           PrintOut = PrintOut & Delimiter & Right$(Str$(j), 1)
       End If
       
       I = I + 1
    Loop

End Function
